# Search

This will use Sphinx for search, with a fallback to Database-powered search if Sphinx is not available.

## Install

Create a new table (This will be created via a migration):

```
CREATE TABLE HS_Sph_Counter
(
    counter_id INTEGER PRIMARY KEY NOT NULL,
    max_doc_id INTEGER NOT NULL
);
```

## Full Text Searching

This encompasses:

* Requests
    * Title and Body (as of 3.2.10 is just body)
    * User may be limited to specific categories, investigate "CATEGORY IN" syntax with Sphinx
* Knowledge Books
* Forums

## Customer Search

If finding Requests by Customer search will use Sphinx is yet to be determined.

* Customer ID (Arbitrarily used by Staff)
* First Name
* Last Name
* Email
* Phone
* Contacted Via (A qualifier rather than query for search )

## Usage

```php
$inquirer = new HS\Search\Sphinx\DbalQuery( $app->make('db.sphinx') );
$search = new HS\Search\SphinxSearch( $inquirer );

$results = $search->requests('some search query');

var_dump($results);

/*
array(20) {
  [0]=>
  array(3) {
    ["id"]=>
    string(6) "135837"
    ["xrequest"]=>
    string(5) "37712"
    ["xcategory"]=>
    string(1) "6"
  }
  [1]=>
  array(3) {
    ["id"]=>
    string(6) "135829"
    ["xrequest"]=>
    string(5) "37712"
    ["xcategory"]=>
    string(1) "6"
  }
  ...
*/
```

You'll notice that `$app->make('db.sphinx')` is used. That example used Doctrine's Dbal connection object to run the MySQL-like queries:

```php
$app->singleton('db.sphinx', function()
{
    $config = new \Doctrine\DBAL\Configuration();

    $connectionParams = array(
        'host' => '127.0.0.1',
        'driver' => $driver,
        'port' => '9306',
        'charset' => 'UTF8'
    );

    return \Doctrine\DBAL\DriverManager::getConnection($connectionParams, $config);
});

$search = $app->make('search');
$results = $search->requests($_GET['q']);

// Some view logic here
$viewPresenter = new HS\View\Search\FulltextSearchPresenter($results, $app['view']);
$out = $viewPresenter->render();
```

Behind the scenes in the Search library:

1. `$app['search']` resolves to `HS\Search\SphinxSearch`
2. `SphinxSearch` uses `HS\Search\Sphinx\DbalQuery` (which uses `$app['db.sphinx']`)
3 `DbalQuery` returns and instance of `HS\Search\Sphinx\SphinxQueryResult`, which parses database IDs from the Sphinx results.
4. Finally `SphinxSearch` uses `HS\Domain\Repo\HistoryItemRepostory` to grab History and it's parent Request

## Request Indexing Strategies

### Keep in Mind:

* Notes are Merged, which don't result in an easy way to say "this request" or "these notes" have been updated (in the eyes of Sphinx indexer/sql queries)
* Notes are added to old requests - only the note (HS_Request_History) table entries have dates to test if there are requests with new notes
* These all assume "plain" indexes over "real-time"

### Strategy A: Each Note is Indexed

* Each Note is indexed, and includes **duplicate** entries of Request meta:
    * Customer Data (email, fname, lname, phone)
    * Request Subject (title)

**Pros:**

* Easy to index
* Search results get referenced note in front-end display
* Can easily delta index by `HS_Request_History.dtGMTChange`
    * Grab those which have changed "today" OR by "greater than last auto_increment id"

**Cons:**

* Duplicate data in each index column (Does that skew results?)
* Need to `GROUP BY xRequest` in search query - issue of proper relevance ranking?
* Doesn't capture updates from merging (new notes aren't created)
* Doesn't capture updates to requests customer data (only new requests)

### Strategy B: Each Request is Indexed

* Each Request is indexed
* Notes are squashed into a concatenated field (via `sql_joined_field`)

**Pros:**

* No duplicate indexed data
* Can search Requests directly, vs the indirect "find Note and get its related Request"

**Cons:**

* ORM makes grabbing only the first note in a request difficult (sql overhead)
* Don't get the matched note in results display, just get the first note (We don't know what note actually matched, sine notes are all concatenated into one field).
* Delta index is hard, as we need to grab latest updated Request based on Request_History dates (or largest auto_increment id), which makes for expensive SQL(?)
    * Can't really interact with Request_History table, right?
    * It would index requests more than once in either case (that an issue?)

### Strategy C: 1 Index for Requests and 1 for Notes

We create two indexes. One handles Requests, and includes:

* title/subject
* Customer info (email, fname, lname, phone)
* Faux-Field: note

One handles Notes (History History) and includes:

* The note
* Faux-fields: title/subject, email, fname, lname, phone

**Pros:**

* We avoid having duplicate indexed data (con of strategy A)
* Separate Customer Search available, regardless of note content (which mirrors the UI use case)
* Delta index strategy can index new notes based on auto_increment_id


**Cons:**

* To search two indexes at the same time, their columns need to match, so some fudging of column names will be required (not a huge issue, Sphinx handles it well)
* Need to `GROUP BY xRequest` in search query - issue of proper relevance ranking?
* Doesn't capture updates from merging (new notes aren't created)
* Doesn't capture updates to requests customer data (only new requests)


#### Generic Notes

* Need to index new notes AND any requests which were created from new notes
    * Or, alternatively, re-index any request regardless of note updates
    * Or maybe customer data (request index) just isn't updated until main index goes (assuming we have strategy 3 in place)
* Merging is a weird issue in all three strategies (essentially a change in xRequest ID on the notes, so we'll end up with the same notes corresponding to two+ requests until the main index is updated)

## Real Time Index

Real time indexes need to be actively managed within the application. This means both having and listening to events.

> See the Events library for handling of Events. Currently uses `Illuminate\Events` without any application-level wrapper.

Requests:

* Request History created (during email intake, or when a response is added in the Workspace, but not when log items are added).
* Request Deleted (and therefor its Request History).
* Request Merge (to update Request ID for each Request History item).

Knowledge Books:

* New Book Page
* Updating Book Page
* Deleting Book page

Forum Posts:

* New Forum Post [api.forums.lib.php]
* Updating a Forum Post (Can these be edited - yes)
* Deleting a Forum Post [api.forums.lib.php, pages/forums.posts.php]

Customers

* TBD still

## Sphinx

### Sources:

**Requests**

* AND HS_Request.xCategory IN ( ... )
* Searches *all* Request History
* Fields to Index:
    * HS_Request_History.xRequest as request_id - Attribute
    * HS_Request.sTitle - Indexed text (JOIN)
    * tNote - Indexed text (remove HTML)


**Knowledge Books**

* Searches Pages only
* Fields to Index:
    * HS_KB_Pages.xPage - Attribute
    * HS_KB_Pages.tPage - Indexed text
* Questions:
    * Need book/chapter IDs?
    * What else to search?
        * Books likely not needed
        * Chapters may want to search (Searchable attribute?)


**Forums**

* Fields to Index:
    * HS_Forums_Posts.xPostId - Attribute
    * HS_Forums_Posts.tPost - Indexed text
    * Questions
        * Need to index HS_Forums_Topics.sTopic? (Searchable attribute?)

**Stopwords**

* [Build stop words](http://answers.oreilly.com/topic/2746-how-to-handle-stop-words-and-short-words-when-working-with-sphinx/) based on test databases.
* Enable n-gram `ngram_len` for CJK (Chinese, Japanese, Korean) support.

**Match Mode**

Phrase may be appropriate in more cases than "words", but perhaps the option in the UI to switch between word vs match mode would make sense.

* Match modes in [SphinxQL](http://sphinxsearch.com/blog/2010/03/23/porting-legacy-matching-modes-to-sphinxql/)

**Feature Flag**

Investigate feature flag ability. Would be good to turn off search, and switch between DB and Sphinx. If DB instead of Sphinx, then the option to search phrase vs word would need to disappear.

**Server**

Sphinx needs a watcher, such as supervisord, upstart, etc, to keep searchd running. Is that possible to do in a standard install?

Default listens on 2 ports - one for SphinxQL and one for Sphinx API. Likely only need SphinxQL port to be listened on.

### Commands

```
# Index all indexes defined in sphinx.conf
## HS Requests collected in 42-60 seconds, 130212 docs, 92.6 MB
## HS KB       collected in .1-.2 seconds, 289    docs, 0.5  MB
## HS Forums   collected in .5-.6 seconds, 6385   docs, 2.1  MB
sudo indexer --config /etc/sphinxsearch/sphinx.conf --all --rotate

## Run a specific index only (A delta index in this case)
sudo indexer --config /etc/sphinxsearch/sphinx.conf --rotate requests_history_ndx_delta


# Run Searchd Daemon (needs a watch file)
sudo mkdir -p /var/run/sphinxsearch
sudo searchd --config /etc/sphinxsearch/sphinx.conf

sudo searchd --config /etc/sphinx/sphinx.conf --stopwait [or --stop]

sudo searchd --config /etc/sphinx/sphinx.conf --status
```

### Ubuntu Install

Install Sphinx Search:

```
sudo add-apt-repository -y ppa:builds/sphinxsearch-rel21
sudo apt-get update
sudo apt-get install -y sphinxsearch
```

> The `add-apt-repository` command is available via the `software-properties-common` on Ubuntu 14.04 and `python-software-properties` package for versions under 14.04.
> Use `sudo apt-get install -y software-properties-common` or `sudo apt-get install -y python-software-properties` as needed

Configure Sphinx Search:

Edit the `config/search.php` file if needed. The default settings are likely OK, unless you are running Sphinx on a separate server:

```php
<?php
return array(

    /**
     * Enable Real-Time indexing over
     * regular interval indexing
     *
     * Real-Time not currently full supported.
     */
    "realtime" => false,

    /**
     * Sphinx Client Connection
     */
    "connection" => array(
        "driver"  => "pdo_mysql",
        "host"    => "127.0.0.1",
        "port"    => "9306",
        "charset" => "UTF8"
    ),
```

HelpSpot has a command to create the SphinxSearch configuration file. It will save it to the `data` directory.

> You must have valid database credentials set in your `config.php` file for this configuration to be created properly.

```bash
php hs search:config
```

Once that's complete, we need to symlink this configuration file to where Sphinx expects the configuration to be - `/etc/sphinxsearch/sphinx.conf`

```bash
# Move default sphinx.conf file:
sudo mv /etc/sphinxsearch/sphinx.conf /etc/sphinxsearch/sphinx-orig.conf

# Symlink HelpSpot's configuration file to sphinx.conf
sudo ln -s /vagrant/HelpSpot/data/sphinx.conf /etc/sphinxsearch/sphinx.conf
```

Start the SphinxSearch service:

```bash
# Create this directory if it does not exist
sudo mkdir /var/run/sphinxsearch

# Start Sphinx
sudo service sphinxsearch start
```

Index the database:

```bash
sudo indexer --all --rotate
```

Indexing should often take under a minute, but will depend upon the size of your database including Requests (and history), knowledge base and forums usage and size.
